# coding: utf-8
from celery.app import app_or_default
from celery.app.control import Control
from celery_balancer import logger


class CeleryApi(object):

    # celery добавляет свой префикс к именам воркеров
    WORKER_NAME_PREFIX = u'celery@'

    def __init__(self):
        self.control = Control(app_or_default())
        self.inspect = self.control.inspect()

    def get_launched_workers_names(self):
        """
        Получение списка имен запущенных воркеров.
        :return: список имен запущенных воркеров
        :rtype : list
        """
        launched_workers = self.__call_inspect_method(u'stats')
        launched_worker_names = launched_workers.keys()
        return [self.__delete_celery_prefix(x) for x in launched_worker_names]

    def get_active_proc_cnt(self):
        """
        Полученеие количества обработчкиов, занятых работой.
        :return: {имя воркерв: кол-во, ...}
        """
        res = dict()
        active_workers_dict = self.__call_inspect_method(u'active')
        for worker_name, active_procs in active_workers_dict.items():
            res[self.__delete_celery_prefix(worker_name)] = len(active_procs)
        return res

    def get_total_proc_cnt(self):
        """
        Получение общего числа обработчиков.
        :return: {имя воркера: кол-во, ...}
        """
        res = dict()
        workers_info_dict = self.__call_inspect_method(u'stats')
        for worker_name, worker_info in workers_info_dict.items():
            res[self.__delete_celery_prefix(worker_name)] = (
                worker_info['pool']['writes']['inqueues']['total']
            )
        return res

    def get_reserved_tasks_cnt(self):
        """
        Количество заданий, которые воркер уже взял из очереди,
        но еще не начал выполнять.
        :return: {имя воркера: кол-во, ...}
        """
        res = dict()
        reserved_tasks_dict = self.__call_inspect_method(u'reserved')
        for worker_name, reserved_tasks in reserved_tasks_dict.items():
            res[self.__delete_celery_prefix(worker_name)] = len(reserved_tasks)
        return res

    def grow_proc_cnt(self, n, worker_name):
        """Увеличение кол-ва обработчиков у воркера"""
        self.control.pool_grow(n, [self.__add_celery_prefix(worker_name)])

    def shrink_proc_cnt(self, n, worker_name, reply=False):
        """Уменьшение кол-ва обработчиков у воркера"""
        worker_name = self.__add_celery_prefix(worker_name)
        res = self.control.pool_shrink(
            n, [worker_name], reply=reply
        )
        if reply:
            return res and res[0][worker_name].get('ok', '') or ''

    def bind_wokrkers_to_queue(self, workers_names, queue_name):
        """
        Привязка воркеров к очереди - воркер начнет брать из нее задания.
        :param workers_names: список имен воркеров
        :param queue_name: имя очереди
        """
        assert isinstance(workers_names, (list, tuple))
        workers_names = [self.__add_celery_prefix(x) for x in workers_names]
        self.control.add_consumer(destination=workers_names, queue=queue_name)

    def unbind_worker_from_queue(self, worker_name, queue_name):
        """
        Отвязка воркера от очереди - воркер перестанет брать из нее задания.
        """
        worker_name = self.__add_celery_prefix(worker_name)
        self.control.cancel_consumer(destination=[worker_name],
                                     queue=queue_name)

    def __add_celery_prefix(self, worker_name):
        return ''.join([self.WORKER_NAME_PREFIX, worker_name])

    def __delete_celery_prefix(self, worker_name):
        return worker_name.replace(self.WORKER_NAME_PREFIX, u'')

    def __call_inspect_method(self, method_name):
        """
        Вызов методов класса Inspect.
        Обрабатывает пустой результат и ошибку в librabbitmq.
        В librabbitmq есть неверное поведение на определенных данных, в
        результате которого валится "ConnectionError: Bad frame read".
        Они это пофиксили, но забыли включить в версию 1.6.1.
        Видимо включат в 1.6.2.
        Источники:
        https://github.com/celery/librabbitmq/issues/30
        https://github.com/celery/librabbitmq/issues/48
        https://github.com/celery/librabbitmq/pull/59
        :param method_name: имя метода
        :rtype : dict
        """
        from librabbitmq import ConnectionError
        try:
            res = getattr(self.inspect, method_name)() or dict()
        except ConnectionError:
            res = dict()
            logger.info(u'Suppressed ConnectionError')
        return res
