# coding: utf-8
import datetime
import decimal
from celery_balancer import logger
from celery_balancer.helpers import nested_commit_on_success
from celery_balancer.models import Worker


class DecisionMaker(object):

    def __init__(self):
        self.rounding = lambda x: x.quantize(1, rounding=decimal.ROUND_HALF_UP)

    def run(self, info):
        """
        Принятие решения о количестве обработчиков для каждого воркера.
        :param info:
            info = {
                'workers': {
                    'worker_name_1': {
                        'queue_name': str,
                        'queue_priority': number,
                        'queue_is_active': bool,
                        'active_proc_cnt': number,
                        'total_proc_cnt': number,
                        'queued_tasks_cnt': number,
                        'decision_interval': number,
                        'last_decision_datetime': datetime,
                    },
                    'worker_name_2': {...},
                    ...,
                }.
                'servers': {
                    server_1: {
                        'workers_list': ['worker_name_1', ...],
                        'cur_metrics': {
                            'cur_metric_name_1': metric_1_measure,
                            'cur_metric_name_2': metric_2_measure,
                            ...,
                        },
                        'max_metrics': {
                            'max_metric_name_1': metric_1_measure,
                            'max_metric_name_2': metric_2_measure,
                            ...,
                        },
                    },
                    server_2: {...},
                    ...,
                }
            }
        :return:
            decision = {
                'worker_name_1': {
                    'cur_proc_cnt': number,
                    'new_proc_cnt': number,
                    'force': bool,
                },
                ...,
            }
        """
        decision = dict()
        workers_from_overloaded_servers = (
            self.__get_workers_from_overloaded_servers(info[u'servers'])
        )

        # Ворекры, для которых будем решать понижать ко-во обработчиков,
        # соблюдая приоритет.
        workers_to_shrink_by_priority = []

        # Воркеры, которым будем решать выставить 0 обработчиков.
        workers_to_set_0_proc_cnt = []

        # Воркеры, которым будем решать выставить 0 обработчиков, force=True.
        workers_to_force_set_0_proc_cnt = []

        # Воркеры, которым будем решать выставить кол-во обработчиков так,
        # чтобы не осталось простаивающих.
        workers_to_cut_free_procs = []

        # Воркеры, которым будем решать выставить кол-во обработчиков
        # в соотв-вии с приоритетом.
        workers_to_set_proc_cnt_by_priority = []

        for worker_name, worker_info in info[u'workers'].items():
            # Исключаем воркеров, у которых нет заданий в очередеи и
            # нет обработчиков.
            if (
                worker_info[u'total_proc_cnt'] == 0 and
                worker_info[u'queued_tasks_cnt'] == 0
            ):
                continue
            if not worker_info[u'queue_is_active']:
                workers_to_force_set_0_proc_cnt.append(worker_name)
            elif (
                worker_info[u'active_proc_cnt'] == 0 and
                worker_info[u'total_proc_cnt'] > 0
            ):
                workers_to_set_0_proc_cnt.append(worker_name)
            elif worker_name in workers_from_overloaded_servers:
                workers_to_shrink_by_priority.append(worker_name)
            elif (
                worker_info[u'active_proc_cnt'] > 0 and
                worker_info[u'queued_tasks_cnt'] == 0
            ):
                workers_to_cut_free_procs.append(worker_name)
            else:
                workers_to_set_proc_cnt_by_priority.append(worker_name)

        decision.update(self.__get_shrink_by_priority_decision(
            workers_to_shrink_by_priority, info[u'workers']
        ))
        decision.update(self.__get_set_0_proc_decision(
            workers_to_set_0_proc_cnt, info[u'workers']
        ))
        decision.update(self.__get_set_0_proc_decision(
            workers_to_force_set_0_proc_cnt, info[u'workers'], force=True,
        ))
        decision.update(self.__get_cut_free_procs_decision(
            workers_to_cut_free_procs, info[u'workers']
        ))
        decision.update(self.__get_proc_cnt_by_priority_decision(
            workers_to_set_proc_cnt_by_priority, info[u'workers']
        ))

        logger.info(u'Decision: %s' % decision)
        return decision

    @staticmethod
    def __get_workers_from_overloaded_servers(servers_info):
        """
        Получение имен воркеров, относящихся к перегруженным серверам.
        :return: workers_names_set
        """
        workers_from_overloaded_servers = set()
        for server_name, server_info in servers_info.items():
            server_overloaded = False
            for cur_metric_name, cur_metric_value in server_info[
                u'cur_metrics'
            ].items():
                max_metric_name = cur_metric_name.replace(u'cur', u'max')
                max_metric_value = server_info[u'max_metrics'][max_metric_name]
                if (
                    cur_metric_value is not None and
                    max_metric_value is not None and
                    cur_metric_value >= max_metric_value
                ):
                    server_overloaded = True
                    break
            if server_overloaded:
                workers_from_overloaded_servers.update(
                    server_info[u'workers_list']
                )
        return workers_from_overloaded_servers

    @staticmethod
    def __get_set_0_proc_decision(workers_list, workers_info, force=False):
        """Решение выставить 0 обработчиков."""
        decision = dict()
        for worker_name in workers_list:
            decision[worker_name] = {
                u'cur_proc_cnt': workers_info[worker_name][u'total_proc_cnt'],
                u'new_proc_cnt': 0,
                u'force': force,
            }
        return decision

    @staticmethod
    def __get_cut_free_procs_decision(workers_list, workers_info):
        """
        Решение понизить кол-во обработчиков до числа активных,
        то есть убрать свободных от работы обработчиков.
        """
        decision = dict()
        for worker_name in workers_list:
            decision[worker_name] = {
                u'cur_proc_cnt': workers_info[worker_name][u'total_proc_cnt'],
                u'new_proc_cnt': workers_info[worker_name][u'active_proc_cnt'],
            }
        return decision

    # TODO: у этого метода много общего с __get_shrink_by_priority_decision,
    # возможно стоит их объединить.
    def __get_proc_cnt_by_priority_decision(self, workers_list, workers_info):
        """Решение выставить кол-во обработчиков в соотв-вии с приоритетами."""
        decision = dict()

        workers_list = self.__exclude_workers_by_decision_interval(
            workers_list, workers_info
        )

        if not workers_list:
            return decision

        workers_sorted_by_priority = sorted(
            workers_list,
            key=lambda x: workers_info[x][u'queue_priority'],
        )
        # Лидирующим воркером является воркер с наивысшим приоритетом у очереди
        lider_worker_name = workers_sorted_by_priority[0]
        other_workers = workers_sorted_by_priority[1:]

        lider_priority = workers_info[lider_worker_name][u'queue_priority']
        lider_cur_proc_cnt = (
            workers_info[lider_worker_name][u'total_proc_cnt'] or 1
        )

        proc_cnt_was_growed = False
        for worker_name in other_workers:
            worker_priority = workers_info[worker_name][u'queue_priority']
            worker_cur_proc_cnt = workers_info[worker_name][u'total_proc_cnt']
            # Отсюда выражаем worker_new_proc_cnt:
            # lider_priority / worker_priority =
            # worker_new_proc_cnt / lider_cur_proc_cnt
            worker_new_proc_cnt = self.rounding(
                lider_cur_proc_cnt * (lider_priority /
                                      decimal.Decimal(worker_priority))
            )
            worker_new_proc_cnt, proc_cnt_was_growed = self.__process_decision(
                worker_new_proc_cnt, worker_cur_proc_cnt,
                worker_name, proc_cnt_was_growed, grow=True,
            )
            if worker_new_proc_cnt < worker_cur_proc_cnt:
                # Нельзя понижать кол-во обработчиков.
                worker_new_proc_cnt = worker_cur_proc_cnt
            decision[worker_name] = {
                u'cur_proc_cnt': worker_cur_proc_cnt,
                u'new_proc_cnt': worker_new_proc_cnt,
            }

        lider_cur_proc_cnt = workers_info[lider_worker_name][u'total_proc_cnt']
        # Если никому не повысили кол-во обработчиков, то повышаем лидеру.
        if not proc_cnt_was_growed:
            lider_new_proc_cnt = lider_cur_proc_cnt + 1
        else:
            lider_new_proc_cnt = lider_cur_proc_cnt or 1

        lider_new_proc_cnt, _ = self.__process_decision(
            lider_new_proc_cnt, lider_cur_proc_cnt,
            lider_worker_name, True, grow=True,
        )

        decision[lider_worker_name] = {
            u'cur_proc_cnt': lider_cur_proc_cnt,
            u'new_proc_cnt': lider_new_proc_cnt,
        }
        return decision

    def __get_shrink_by_priority_decision(self, workers_list, workers_info):
        """Решение понизить кол-во обработчиков, соблюдая приоритет."""
        decision = dict()

        # исключаем воркеров с 0 обработчиков.
        workers_list = filter(lambda x: workers_info[x][u'total_proc_cnt'] > 0,
                              workers_list)

        workers_list = self.__exclude_workers_by_decision_interval(
            workers_list, workers_info
        )

        if not workers_list:
            return decision

        workers_sorted_by_priority_desc = sorted(
            workers_list,
            key=lambda x: workers_info[x][u'queue_priority'],
            reverse=True,
        )

        # Лидирующим воркером является воркер
        # с наименьшим приоритетом у очереди.
        lider_worker_name = workers_sorted_by_priority_desc[0]
        other_workers = workers_sorted_by_priority_desc[1:]

        lider_priority = workers_info[lider_worker_name][u'queue_priority']
        lider_cur_proc_cnt = workers_info[lider_worker_name][u'total_proc_cnt']

        proc_cnt_was_shrinked = False
        for worker_name in other_workers:
            worker_priority = workers_info[worker_name][u'queue_priority']
            worker_cur_proc_cnt = workers_info[worker_name][u'total_proc_cnt']
            # Отсюда выражаем worker_new_proc_cnt:
            # lider_priority / worker_priority =
            # worker_new_proc_cnt / lider_cur_proc_cnt
            worker_new_proc_cnt = self.rounding(
                lider_cur_proc_cnt * (lider_priority /
                                      decimal.Decimal(worker_priority))
            )
            worker_new_proc_cnt, proc_cnt_was_shrinked = (
                self.__process_decision(
                    worker_new_proc_cnt, worker_cur_proc_cnt,
                    worker_name, proc_cnt_was_shrinked,
                    grow=False,
                )
            )
            if worker_new_proc_cnt > worker_cur_proc_cnt:
                # Нельзя повышать кол-во обработчиков.
                worker_new_proc_cnt = worker_cur_proc_cnt

            decision[worker_name] = {
                u'cur_proc_cnt': worker_cur_proc_cnt,
                u'new_proc_cnt': worker_new_proc_cnt,
            }

        # Если никому не понизили кол-во обработчиков, то понижаем лидеру.
        if not proc_cnt_was_shrinked:
            lider_new_proc_cnt = lider_cur_proc_cnt - 1
        else:
            lider_new_proc_cnt = lider_cur_proc_cnt

        lider_new_proc_cnt, _ = self.__process_decision(
            lider_new_proc_cnt, lider_cur_proc_cnt,
            lider_worker_name, True, grow=False,
        )

        decision[lider_worker_name] = {
            u'cur_proc_cnt': lider_cur_proc_cnt,
            u'new_proc_cnt': lider_new_proc_cnt,
        }
        return decision

    @staticmethod
    @nested_commit_on_success
    def __process_decision(
            worker_new_proc_cnt, worker_cur_proc_cnt, worker_name,
            proc_cnt_was_changed, grow=True
    ):
        if worker_new_proc_cnt != worker_cur_proc_cnt:
            if (
                (grow and worker_new_proc_cnt > worker_cur_proc_cnt) or
                (not grow and worker_new_proc_cnt < worker_cur_proc_cnt)
            ):
                proc_cnt_was_changed = True
            Worker.objects.set_fields_values(
                {u'last_decision_datetime':
                 datetime.datetime.now()},
                (worker_name, )
            )
        return worker_new_proc_cnt, proc_cnt_was_changed

    @staticmethod
    def __exclude_workers_by_decision_interval(workers_list, workers_info):
        """
        Исключаем из списка воркеров те, для которых
        еще не настало время принимать решение.
        """
        filtered_workers_list = []
        for worker_name in workers_list:
            last_decision_datetime = workers_info[
                worker_name
            ][u'last_decision_datetime']
            decision_interval = workers_info[
                worker_name
            ][u'decision_interval']
            if (
                not last_decision_datetime or
                datetime.datetime.now() >
                last_decision_datetime +
                datetime.timedelta(seconds=decision_interval)
            ):
                filtered_workers_list.append(worker_name)
            else:
                logger.info(u'Decision interval '
                            u'not expired yet for worker "%s".' % worker_name)
        return filtered_workers_list