# coding: utf-8
from celery_balancer import logger
from celery_balancer.models import Server, Worker


class InfoGetter(object):

    def __init__(self, celery_api, mq_api, metrics_api):
        self.celery_api = celery_api
        self.mq_api = mq_api
        self.metrics_api = metrics_api

    def run(self):
        """
        Сбор информации о серверах, очередях и воркерах.
        :return:
            info = {
                'workers': {
                    'worker_name_1': {
                        'queue_name': str,
                        'queue_priority': number,
                        'queue_is_active': bool,
                        'active_proc_cnt': number,
                        'total_proc_cnt': number,
                        'queued_tasks_cnt': number,
                        'decision_interval': number,
                        'last_decision_datetime': datetime,
                    },
                    'worker_name_2': {...},
                    ...,
                }.
                'servers': {
                    server_1: {
                        'workers_list': ['worker_name_1', ...],
                        'cur_metrics': {
                            'cur_metric_name_1': metric_1_measure,
                            'cur_metric_name_2': metric_2_measure,
                            ...,
                        },
                        'max_metrics': {
                            'max_metric_name_1': metric_1_measure,
                            'max_metric_name_2': metric_2_measure,
                            ...,
                        },
                    },
                    server_2: {...},
                    ...,
                }
            }
        """
        info = {
            u'workers': self.__get_workers_info(),
            u'servers': self.__get_servers_info(),
        }
        logger.info(u'Info: %s' % info)
        return info

    def __get_servers_info(self):
        res = dict()
        workers_names = Server.objects.get_workers_names_affect_on_servers()
        max_metrics_names = [u''.join([u'max_', x])
                             for x in self.metrics_api.metrics]
        max_metrics = Server.objects.get_fields_values(max_metrics_names)
        for server_name in Server.objects.get_names():
            server_info = {
                u'workers_list': workers_names[server_name],
                u'cur_metrics': dict(),
                u'max_metrics': max_metrics[server_name],
            }

            for metric in self.metrics_api.metrics:
                max_metric_name = u''.join([u'max_', metric])
                cur_metric_name = u''.join([u'cur_', metric])
                # Будем получать текущие метрики, если заданы максимальные
                if server_info[u'max_metrics'][max_metric_name] is not None:
                    server_info[u'cur_metrics'][cur_metric_name] = (
                        self.metrics_api.get_metric(metric, server_name)
                    )
                else:
                    server_info[u'cur_metrics'][cur_metric_name] = None
            res[server_name] = server_info
        return res

    def __get_workers_info(self):
        res = dict()
        launched_workers_names = set(
            self.celery_api.get_launched_workers_names()
        )
        all_workers_names = set(Worker.objects.get_names())
        workers_names_to_get_info = all_workers_names.intersection(
            launched_workers_names
        )
        active_proc_cnt = self.celery_api.get_active_proc_cnt()
        total_proc_cnt = self.celery_api.get_total_proc_cnt()
        for worker_name in workers_names_to_get_info:
            worker = Worker.objects.get_by_name(worker_name)
            queue = worker.queue
            res[worker_name] = {
                u'queue_name': queue.name,
                u'queue_priority': queue.priority,
                u'queue_is_active': queue.is_active,
                u'active_proc_cnt': active_proc_cnt.get(worker_name, 0),
                u'total_proc_cnt': total_proc_cnt.get(worker_name, 0),
                u'queued_tasks_cnt': self.mq_api.get_tasks_cnt(queue.name),
                u'decision_interval': queue.decision_interval,
                u'last_decision_datetime': worker.last_decision_datetime,
            }

        return res
