# coding: utf-8
from django.core.exceptions import FieldError
from django.db.models import Manager, Q, Min
from ordereddict import OrderedDict
from celery_balancer.helpers import nested_commit_on_success


class BaseManager(Manager):

    def get_names(self):
        return tuple(self.all().values_list('name', flat=True))

    def get_settings_names(self):
        res = OrderedDict()
        for field in filter(
                lambda x: x.name in self.model.settings_fields_with_comments,
                self.model._meta.local_fields
        ):
            res[field.name] = '. '.join([
                field.verbose_name,
                self.model.settings_fields_with_comments[field.name]
            ])
        return res

    def set_fields_values(self, fields_values, obj_names_list=None):
        assert isinstance(fields_values, dict)
        q = Q()
        if obj_names_list:
            assert isinstance(obj_names_list, (list, tuple, set))
            q &= Q(name__in=obj_names_list)
        self.filter(q).update(**fields_values)

    def get_fields_values(self, fields_name_list=None, obj_names_list=None):
        """:rtype : dict: {obj_name: {field: value}}"""
        q = Q()
        if obj_names_list:
            assert isinstance(obj_names_list, (list, tuple, set))
            q &= Q(name__in=obj_names_list)

        if fields_name_list:
            assert isinstance(fields_name_list, (list, tuple, set))
        else:
            fields_name_list = (
                self.model.settings_fields_with_comments.keys()
            )

        res = dict()
        values = ['name'] + list(fields_name_list)
        try:
            objs = self.filter(q).order_by('name').values(*values)
        except FieldError:
            raise Exception(u'Неверное имя настройки.')
        for obj in objs:
            name = obj['name']
            del obj['name']
            res[name] = obj
        return res

    def delete_by_names(self, name_list):
        assert isinstance(name_list, (list, tuple, set))
        self.filter(name__in=name_list).delete()

    def exists_by_name(self, name):
        return self.filter(name=name).exists()

    def get_by_name(self, name):
        return self.get(name=name)


class QueueManager(BaseManager):
    """Менеджер работы с моделью Queue"""

    def get_min_decision_interval(self):
        return self.aggregate(
            Min('decision_interval')
        )['decision_interval__min']

    def get_queues_workers_names(self, queue_names_list=None):
        """Получение имен воркеров очередей.
        :rtype : dict: {queue_name: (workers_names_tuple)}
        """
        q = Q()
        if queue_names_list:
            assert isinstance(queue_names_list, (list, tuple, set))
            q &= Q(name__in=queue_names_list)
        queues = self.filter(q)

        res = dict()
        for queue in queues:
            res[queue.name] = tuple(queue.get_workers_names())

        return res


class ServerManager(BaseManager):
    """Менеджер работы с моделью Server"""

    def register_server(self, name, host, username, password):
        if self.filter(name=name).exists():
            raise Exception(u'Сервер с таким именем уже зарегистрирован.')
        self.create(
            name=name,
            host=host,
            username=username,
            password=self.model.encrypt(password),
        )

    def cancel_servers_registration(self, server_name_list):
        self.delete_by_names(server_name_list)

    def set_fields_values(self, fields_values, server_name_list=None):
        assert isinstance(fields_values, dict)
        if 'password' in fields_values:
            fields_values['password'] = self.model.encrypt(
                fields_values['password']
            )
        q = Q()
        if server_name_list:
            assert isinstance(server_name_list, (list, tuple, set))
            q &= Q(name__in=server_name_list)
        self.filter(q).update(**fields_values)

    def get_workers_names_affect_on_servers(self, server_name_list=None):
        """Получение имен воркеров, влияющих на загруженность серверов
        :rtype : dict: {server_name: (workers names tuple)}
        """
        q = Q()
        if server_name_list:
            assert isinstance(server_name_list, (list, tuple, set))
            q &= Q(name__in=server_name_list)
        servers = self.filter(q)

        res = dict()
        for server in servers:
            res[server.name] = tuple(
                server.workers.values_list('name', flat=True)
            )
        return res


class WorkerManager(BaseManager):
    """Менеджер работы с моделью Worker"""

    def create_workers_for_queue(self, queue, workers_names_list=None):
        for worker_name in workers_names_list:
            self.create(
                name=worker_name,
                queue=queue,
            )


class ServerWorkerM2MManager(Manager):
    """Менеджер работы с моделью ServerWorkerM2M"""

    @nested_commit_on_success
    def set_servers_affected_by_worker(self, affection_dict):
        """
        Задание на какие сервера влияет воркер.
        :param affection_dict: {имя воркера: (список имен серверов), ...}
        :return: Success/traceback
        """
        from celery_balancer.models import Server, Worker
        assert isinstance(affection_dict, dict)
        for worker_name, servers_names in affection_dict.items():
            assert isinstance(servers_names, (list, tuple, set))
            worker = Worker.objects.get_by_name(worker_name)
            self.delete_by_worker(worker)
            for server_name in servers_names:
                server = Server.objects.get_by_name(server_name)
                self.create(
                    server=server,
                    worker=worker,
                )

    def delete_by_worker(self, worker):
        self.filter(worker=worker).delete()