# coding: utf-8
import re
import traceback
import decimal
from celery_balancer import logger
from celery_balancer.models import Server


class LinuxOsApi(object):
    """
    Клссс для комманд ОС linux
    """

    def __init__(self):
        # Маппинг {имя сервера: ssh соединение, ...}
        self.ssh_connetctions = dict()

        servers_names = Server.objects.get_names()
        for server_name in servers_names:
            self.ssh_connetctions[server_name] = (
                self.create_ssh_connection(server_name)
            )

    @staticmethod
    def create_ssh_connection(server_name):
        """
        Создание ssh соединения с серверов.
        :param server_name: иям сервера
        """
        import paramiko
        server_settings = Server.objects.get_fields_values(
            ('host', 'username', 'password'),
            (server_name, ),
        )[server_name]
        server_settings['password'] = Server.decrypt(
            server_settings['password']
        )
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(
            server_settings['host'],
            username=server_settings['username'],
            password=server_settings['password'],
        )
        return ssh

    def exec_command(self, server_name, command, reconnect=True):
        """
        Выполнение команды на удаленном сервере.
        :param server_name: имя сервера
        :param command: комманда
        :param reconnect: будет ли попытка переподключения к серверу
                          в случае потери соединения
        :return: результат выполнения команды command
        """
        try:
            stdin, stdout, stderr = (
                self.ssh_connetctions[server_name].exec_command(command)
            )
        except Exception:
            logger.error(traceback.format_exc())
            if reconnect:
                self.ssh_connetctions[server_name] = (
                    self.create_ssh_connection(server_name)
                )
                self.exec_command(server_name, command, reconnect=False)
            else:
                return u''
        else:
            return stdout.read()

    def get_mem(self, server_name):
        total_mem = self.__get_proc_meminfo_data(u'MemTotal', server_name)
        if total_mem is None:
            return None

        free_mem = self.__get_proc_meminfo_data(u'MemFree', server_name)
        if free_mem is None:
            return None

        buffers = self.__get_proc_meminfo_data(u'Buffers', server_name)
        if buffers is None:
            return None

        cached = self.__get_proc_meminfo_data(u'Cached', server_name)
        if cached is None:
            return None

        swap_cached = self.__get_proc_meminfo_data(u'SwapCached', server_name)
        if swap_cached is None:
            return None

        return total_mem - (free_mem + buffers + cached + swap_cached)

    def get_swp(self, server_name):
        total_swap = self.__get_proc_meminfo_data(u'SwapTotal', server_name)
        if total_swap is None:
            return None

        free_swap = self.__get_proc_meminfo_data(u'SwapFree', server_name)
        if free_swap is None:
            return None

        return total_swap - free_swap

    def get_load_avg_1(self, server_name):
        command = u'uptime'
        res = self.exec_command(server_name, command)
        if not res:
            return None
        res = res.strip(u'\n')
        _, load_avg = res.split(u'load average:')
        load_avg = load_avg.strip()
        load_avg_1, _, _ = load_avg.split(u', ')
        return decimal.Decimal(load_avg_1)

    def get_load_avg_5(self, server_name):
        command = u'uptime'
        res = self.exec_command(server_name, command)
        if not res:
            return None
        res = res.strip(u'\n')
        _, load_avg = res.split(u'load average:')
        load_avg = load_avg.strip()
        _, load_avg_5, _ = load_avg.split(u', ')
        return decimal.Decimal(load_avg_5)

    def get_load_avg_15(self, server_name):
        command = u'uptime'
        res = self.exec_command(server_name, command)
        if not res:
            return None
        res = res.strip(u'\n')
        _, load_avg = res.split(u'load average:')
        load_avg = load_avg.strip()
        _, _, load_avg_15 = load_avg.split(u', ')
        return decimal.Decimal(load_avg_15)

    def __get_proc_meminfo_data(self, data_name, server_name):
        """Получение данных из /proc/meminfo"""
        command = u'grep %s /proc/meminfo -w' % data_name
        raw = self.exec_command(server_name, command)
        if not raw:
            return None
        raw = raw.strip(u'\n')
        raw = raw.strip()
        _, data_kb, _ = re.split(ur'\s+', raw)
        data_mb = int(data_kb) / 1024
        return data_mb